local addonName, addonTable = ...
local debug = false
--[===[@debug@
debug = true
--@end-debug@]===]


local strfind, strformat, strgsub, strlower, strmatch, tostring = string.find, string.format, string.gsub, string.lower, strmatch, tostring
local _, date, ipairs, next, pairs, select, time, tinsert, tsort, type, wipe = _, date, ipairs, next, pairs, select, time, table.insert, table.sort, type, wipe

-- LibDataBroker
local ldb = LibStub:GetLibrary("LibDataBroker-1.1", true)
local PTLDB
local LDBIcon = ldb and LibStub("LibDBIcon-1.0",true)
local LibQTip = LibStub('LibQTip-1.0')
local PT = LibStub("AceAddon-3.0"):NewAddon(addonName, "AceConsole-3.0", "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale(addonName, false)
local currentRealmChar, key, id


local RAID_CLASS_COLORS_FONTS = {}


local function setColor(info, r, g, b, a)
  PT.db.profile[info[#info]] = { r = r, g = g, b = b, a = a }
end

local function getColor(info)
  return PT.db.profile[info[#info]].r, PT.db.profile[info[#info]].g, PT.db.profile[info[#info]].b, PT.db.profile[info[#info]].a
end


local defaults = {
  profile = {
    usedcd = { r=1, g=0, b=0, a=1 },
    freecd = { r=0, g=1, b=0, a=1 },
    useClassColors          = true,
    useCustomClassColors    = true,
    showLDBCooldownsAvailable = true,
    showLDBCooldownsText      = "CD: %d",
    showLDBItemsAvailable     = true,
    showCooldowns = true,
    showBooks     = false,
    minLevel = 1,
    minimapIcon = {
      hide = true,
      minimapPos = 220,
      radius = 80,
    },
  },
  global = {
    roster = {},
    cooldowns = {},
    items = {},
    trackedItems = {
      [108257] = true, -- Truesteel Ingot
      [108996] = true, -- Alchemical Catalyst
      [110611] = true, -- Burnished Leather
      [111366] = true, -- Gearspring Parts
      [111556] = true, -- Hexweave Cloth
      [112377] = true, -- War Paints
      [115524] = true, -- Taladite Crystal
      [115504] = true, -- Fractured Temporal Crystal
    }
  },
}

local professionCooldowns = {
  -- skillLine is the one returned by GetProfessionInfo( ... )
  -- need to check if any of multiple spells is on cooldown
  [171] = { iconid = 108996, spells = { 156587, 168042 }, book = 175880 }, -- alchemy
  [164] = { iconid = 108257, spells = { 171690, 171718 }, book = 176090 }, -- blacksmithing
  [333] = { iconid = 115504, spells = { 169092, 178241 }, book = 177043 }, -- enchating
  [202] = { iconid = 111366, spells = { 169080, 178242 }, book = 177054 }, -- engineering
  [773] = { iconid = 112377, spells = { 169081, 178240 }, book = 177045 }, -- inscription
  [755] = { iconid = 115524, spells = { 170700, 170832 }, book = 176087 }, -- jewelcrafting
  [165] = { iconid = 110611, spells = { 171391, 171713 }, book = 176089 }, -- leatherworking
  [197] = { iconid = 111556, spells = { 168835, 169669 }, book = 176058 }, -- tailoring
}
local professionCooldownSpellIDs = {}

function PT:LevelCheck(info, newMinLevel)
  self.db.profile.minLevel = newMinLevel
end

function PT:OnInitialize() -- ADDON_LOADED, only saved variables are guaranteed to be present
  self.db = LibStub("AceDB-3.0"):New("ProfessionTrackDB", defaults, true)

  LibStub("AceConfig-3.0"):RegisterOptionsTable(addonName, self.GenerateOptions)
  self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonName)

  PTLDB = ldb:NewDataObject(addonName, {
    type = "data source",
    text = addonName,
    icon = "Interface\\Icons\\inv_eng_gearspringparts",
    OnClick = function(clickedframe, button)
      if button == "RightButton" then 
        InterfaceOptionsFrame_OpenToCategory(self.optionsFrame) 
      else
        self:UpdatePlayer()
        self:UpdateItemCounts()
        self:UpdateProfessionCooldowns()
        self:UpdateLDB()
      end
    end,
    OnEnter = function(tt)
      local tooltip = LibQTip:Acquire("PTTooltip", 1, "LEFT") 
      self.tooltip = tooltip
      self:PrepareTooltip(tooltip) 
      tooltip:SmartAnchorTo(tt)
      tooltip:Show()
    end,
    OnLeave = function(tt)
      LibQTip:Release(PT.tooltip)
      self.tooltip = nil
    end,
  })

  LDBIcon:Register(addonName, PTLDB, self.db.profile.minimapIcon)


  currentRealmChar = strformat("%s.%s", GetRealmName(), UnitName("player"))
  if not PT.db.global.cooldowns[currentRealmChar] then
    PT.db.global.cooldowns[currentRealmChar] = {}
  end

  for id, _ in pairs(PT.db.global.trackedItems) do
    -- Pull all tracked items into cache
    GetItemInfo( id )
  end
  
  -- Generate a list of all spellids we want to listen to
  local p, s
  for _, p in pairs(professionCooldowns) do
    for _, s in pairs( p.spells ) do
      professionCooldownSpellIDs[s] = true
    end
    professionCooldownSpellIDs[p.book] = true
  end

  C_Timer.After( 0.1, function() self:GenerateProfessionTextures() end )
end

function PT:GenerateProfessionTextures()
  local done, itemTexture = true
  for _, p in pairs(professionCooldowns) do
    itemTexture = select(10, GetItemInfo( p.iconid ))
    if not itemTexture then
      done = false
    else
      p.texture = "|T"..itemTexture..":16|t"
    end
  end

  -- Repeat if something was not loaded properly
  if not done then
    C_Timer.After( 0.1, function() self:GenerateProfessionTextures() end )
  end
end

function PT:GenerateOptions()
  if not PT.options then

    local function newCounter()
      local i = 0
      return function()
          i = i + 1
          return i
        end
    end
    local count = newCounter()

    PT.options = {
      name = addonName,
      type = 'group',
      handler = PT,
      args = {
        genconfig = {
          name = L["General Settings"],
          type = 'group',
          order = 1,
          get = function(info) return PT.db.profile[info[#info]] end,
          set = function(info, value) PT.db.profile[info[#info]] = value end,
          args = {
            usedcd = {
              name = L["Used CD color"],
              desc = L["Used CD color description"],
              type = 'color',
              order = count(),
              get  = getColor,
              set  = setColor,
              hasAlpha = true,
            },
            freecd = {
              name = L["Free CD color"],
              desc = L["Free CD color description"],
              type = 'color',
              order = count(),
              get  = getColor,
              set  = setColor,
              hasAlpha = true,
            },
            useClassColors = {
              type = "toggle",
              order = count(),
              name = L["Color names by class"],
            },
            useCustomClassColors = {
              type = "toggle",
              order = count(),
              name = L["Use !ClassColors"],
              desc = L["Use !ClassColors addon for class colors used to color the names in the tooltip"],
              get = function(info) return PT.db.profile[info[#info]] end,
              set = function(info, value) 
                PT.db.profile[info[#info]] = value 
                PT:SetupClasscoloredFonts()
              end,
              disabled = function() return not PT.db.profile.useClassColors or not CUSTOM_CLASS_COLORS end,
            },
            showCooldowns = {
              type = "toggle",
              order = count(),
              name = L["Show cooldowns"],
              desc = L["Show columns for profession cooldowns"],
            },
            showBooks = {
              type = "toggle",
              order = count(),
              name = L["Show a 'b' for 'Secref of ...'"],
              desc = L["Adds a 'b' to the cooldown column if the book-spell 'Secret of draenor ...' is on cooldown. (Will only show if the main spell is on cooldown too!)"],
            },
            
            showLDBCooldownsAvailable = {
              type = "toggle",
              order = count(),
              name = L["LDB: Show Cooldowns"],
              desc = L["Add the number of available cooldoewns to the LDB text"],
              set = function(info, value) 
                      PT.db.profile[info[#info]] = value
                      PT:UpdateLDB()
                    end,
            },
            showLDBCooldownsText = {
              type = "input",
              order = count(),
              name = L["LDB: Cooldowns text"],
              desc = L["Formatstring for the text, use %d for the number of cooldowns"],
              set = function(info, value) 
                      PT.db.profile[info[#info]] = value
                      PT:UpdateLDB()
                    end,
            },
            showLDBItemsAvailable = {
              type = "toggle",
              order = count(),
              name = L["LDB: Show Items"],
              desc = L["Add the number of profession items on the current char to the LDB text"],
              set = function(info, value) 
                      PT.db.profile[info[#info]] = value
                      PT:UpdateLDB()
                    end,
            },
            minLevel = {
              type = "range",
              order = count(),
              min = 1;
              max = MAX_PLAYER_LEVEL_TABLE[GetExpansionLevel()];
              step = 1,
              name = L["Minimum Level"],
              desc = L["Characters below this level will not be shown"],
              set = "LevelCheck",
            },
            minimapIcon = {
              type = "toggle",
              name = L["Show minimap button"],
              desc = L["Show the PT minimap button"],
              order = count(),
              get = function(info) return not PT.db.profile.minimapIcon.hide end,
              set = function(info, value)
                PT.db.profile.minimapIcon.hide = not value
                if value then LDBIcon:Show(addonName) else LDBIcon:Hide(addonName) end
              end,
            },
          },
        },
        tracking = {
          name = L["Tracked Items"],
          type = 'group',
          order = 1,
          get = function(info) return PT.db.profile[info[#info]] end,
          set = function(info, value) PT.db.profile[info[#info]] = value end,
          args = {},
        },
      },
    }
    PT.options.args.tracking.args.adddefault = {
      type = "execute",
      name = L["Add Default Items"],
      func = function(info)
        local iname, trackid
        for trackid, _ in pairs(defaults.global.trackedItems) do
          iname = GetItemInfo(trackid)
          if trackid and trackid > 0 and iname and not PT.options.args.tracking.args[tostring(trackid)] then
            PT.db.global.trackedItems[trackid] = true
            PT.options.args.tracking.args[tostring(trackid)]  = {
              type = "execute",
              name = iname,
              func = function(info)
                PT.db.global.trackedItems[tonumber(info[#info])] = false
                PT.options.args.tracking.args[info[#info]] = nil
              end,
            }
          end
        end
      end, 
      order = count()
    }

    PT.options.args.tracking.args.addsavageblood = {
      type = "execute",
      name = L["Add Savage Blood"],
      func = function(info)
        local trackid = 118472
        local iname = GetItemInfo(trackid)
        if trackid and trackid > 0 and iname and not PT.options.args.tracking.args[tostring(trackid)] then
          PT.db.global.trackedItems[trackid] = true
          PT.options.args.tracking.args[tostring(trackid)]  = {
            type = "execute",
            name = iname,
            func = function(info)
              PT.db.global.trackedItems[tonumber(info[#info])] = false
              PT.options.args.tracking.args[info[#info]] = nil
            end,
          }
        end
      end, 
      order = count()
    }

    PT.options.args.tracking.args.itemaddinput = {
      type = "input",
      name = L["Add Item"],
      get = function( info ) return "" end,
      set = function( info, value )
        local trackid = tonumber(value)
        local iname = GetItemInfo(trackid)
        if trackid and trackid > 0 and iname and not PT.options.args.tracking.args[tostring(trackid)] then
          PT.db.global.trackedItems[trackid] = true
          PT.options.args.tracking.args[tostring(trackid)]  = {
            type = "execute",
            name = iname,
            func = function(info)
              PT.db.global.trackedItems[tonumber(info[#info])] = false
              PT.options.args.tracking.args[info[#info]] = nil
            end,
          }
          PT:UpdateItemCounts()
          PT:UpdateLDB()
        end
      end,
      order = count()
    }

    PT.options.args.tracking.args.trackingheader = {
      type = "header",
      name = L["Items"],
      order = count()
    }


    local iname, id, yesno, co
    for id, yesno in pairs(PT.db.global.trackedItems) do
      if yesno then
        iname = GetItemInfo( id )
        PT.options.args.tracking.args[tostring(id)]  = {
          type = "execute",
          name = iname,
          func = function(info)
            PT.db.global.trackedItems[tonumber(info[#info])] = false
            PT.options.args.tracking.args[info[#info]] = nil
          end,
          order = count()
        }
        PT.options.args.tracking.args[tostring(id)..".1"]  = {
          type = "execute",
          name = "^",
          width = "half",
          hidden = (not debug),
          func = function(info)
            print(info[#info], "^")
          end,
          order = count()
        }
        PT.options.args.tracking.args[tostring(id)..".0"]  = {
          type = "execute",
          name = "v",
          width = "half",
          hidden = (not debug),
          func = function(info)
            print(info[#info], "v")
          end,
          order = count()
        }
      end
    end
    PT.options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(PT.db)
  end 
  return PT.options
end


---- Tooltip handling/helpers
local function PrepRosterSlice()
  -- returns a roster filtered by level check into which resets can be added
  local minLevel = PT.db.profile.minLevel
  local roster = {}

  for realmChar, data in pairs(PT.db.global.roster) do
    if minLevel <= data.level then
      local realm, char = strmatch(realmChar, "(.+)%.(.+)")
      roster[realmChar] = { 
        name = char, 
        level = data.level, 
        color = RAID_CLASS_COLORS_FONTS[data.class], 
        items = PT.db.global.items[realmChar], 
        cooldowns = PT.db.global.cooldowns[realmChar],
      }
    end
  end

  -- structure is { realmChar = { (char)name, level, color, items = { name = data }, cooldowns = { name = data } } }
  return roster
end

local function SortRosterSlice(roster)
  -- roster param structure is { realmChar = { (char)name, level, color, items = { name = data }, cooldowns = { name = data } } }
  -- basically, turn the key into sorted tables which point to the value
  local sortedRoster = {}
  local temp = {}
  for realmChar,data in pairs(roster) do
    tinsert(temp, realmChar)
  end
  tsort(temp)

  local curRealm
  for _,realmChar in ipairs(temp) do
    local realm, char = strmatch(realmChar, "(.+)%.(.+)")

    if curRealm ~= realm then
      curRealm = realm
      tinsert(sortedRoster, { name = realm, chars = {} })
    end

    tinsert( sortedRoster[#sortedRoster].chars, roster[realmChar] )
  end

  -- structure is { [iRealm] = { name, chars = { [iChar] = { name, level, color, items = { name = data }, cooldowns = { name = data } } } } }
  return sortedRoster
end


function PT:SetupClasscoloredFonts()
  local class, color, CHOSEN_CLASS_COLORS

  if self.db.profile.useCustomClassColors and CUSTOM_CLASS_COLORS then
    CHOSEN_CLASS_COLORS = CUSTOM_CLASS_COLORS
  else
    CHOSEN_CLASS_COLORS = RAID_CLASS_COLORS
  end
  for class,color in pairs(CHOSEN_CLASS_COLORS) do
    if not RAID_CLASS_COLORS_FONTS[class] then 
      RAID_CLASS_COLORS_FONTS[class] = CreateFont("ClassFont"..class)
      RAID_CLASS_COLORS_FONTS[class]:CopyFontObject(GameTooltipText)
    end
    RAID_CLASS_COLORS_FONTS[class]:SetTextColor(color.r, color.g, color.b)
  end
end

function PT:PrepareTooltip(tooltip)

  --[[  Each char can have up to 2 cooldowns, but 
        any kind of item in the inventroy thanks 
        to garrison buildings

        cd1 cd2 i1  i2  i3  i4  i5  i6  i7  i8
  Char1 [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ]
  Char2 [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ]
  Char3 [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ]
  -----]]--
  local Columns = {}
  local function MyAddColumn(nameType, displayName)
    if nameType then
      if not displayName then
        print("MyAddColumn: missing displayName for", nameType)
        displayName = nameType
      end
      if not formatFunc then
        --formatFunc = CellFormat.Default
      end

      if not Columns[nameType] then
        local colindex = tooltip:AddColumn("CENTER")
        tooltip:SetCell(2, colindex, displayName)

        Columns[nameType] = { col = colindex }
        -- print("adding column", colindex, nameType, displayName)
      else
        print("a column already exists for", nameType)
      end
    end
  end
  local strResetBreaker = "(([^%.]+)%.([^%.]+))%.([^%.]+)"   -- returns realmChar, realm, char, name
  
  local globaldb = self.db.global
  local profiledb = self.db.profile

  local roster = PrepRosterSlice()  -- structure is { realmChar = { (char)name, level, color, resets = { name = data } } }
  local found = {}
  
  tooltip:AddHeader()
  tooltip:AddHeader()
  local trackedItemsSorted = {}

  for id, show in pairs(PT.db.global.trackedItems) do
    if show then
      table.insert( trackedItemsSorted, id )
    end
  end

  table.sort(trackedItemsSorted)
  
  if profiledb.showCooldowns then
    MyAddColumn( "cd1", "CD1")
    MyAddColumn( "cd2", "CD2")
  end

  for _, id in pairs(trackedItemsSorted) do
    --print( currentRealmChar, id, key, GetItemCount( id ) + GetItemCount( id, true) )
    local _, _, _, _, _, _, _, _, _, itemTexture, _ = GetItemInfo( id )
    local formatedTexture = "|T"..itemTexture..":16|t"
    MyAddColumn( id, formatedTexture )
  end





  for dbkey,data in pairs(globaldb.items) do
    local realmChar, realm, char, name = strmatch(dbkey, strResetBreaker)
    if roster[realmChar] then
      found[nameType] = true
    end
  end

  for dbkey,data in pairs(globaldb.cooldowns) do
    local realmChar, realm, char, name = strmatch(dbkey, strResetBreaker)
    if roster[realmChar] then
      found[nameType] = true
    end
  end

  local sortedRoster = SortRosterSlice( roster )
  for _, realm in ipairs(sortedRoster) do

    for _, char in ipairs(realm.chars) do
      -- local charRealm = char.." - "..realm
      local lineName = char.name -- profile.showCharacterRealm and charRealm or char
      local lineFont = profiledb.useClassColors and char.color or nil --  RAID_CLASS_COLORS_FONTS[ self.db.global.charClass[charRealm] ] or nil
      local row = tooltip:AddLine("")

      tooltip:SetCell(row, 1, lineName, lineFont)
      
      for _, itemid in pairs( trackedItemsSorted ) do
        tooltip:SetCell(row, Columns[itemid].col, char.items[itemid] or 0)
      
      end
      local now = time()
      local cdused, cellContent
      if profiledb.showCooldowns then
        local cd = 1
        for skillLine, data in pairs(char.cooldowns) do
          if cd > 2 then
            print("cd > 2, wtf")
          end
          cellContent =  professionCooldowns[skillLine].texture
          if data.reset and data.reset > now then
            if profiledb.showBooks and data.bookReset and data.bookReset > now then
              cellContent = cellContent .."b"
            end
            tooltip:SetCell(row, Columns["cd"..cd].col, cellContent)
            tooltip:SetCellColor(row, Columns["cd"..cd].col, profiledb.usedcd.r, profiledb.usedcd.g, profiledb.usedcd.b, profiledb.usedcd.a)
          else
            tooltip:SetCell(row, Columns["cd"..cd].col, cellContent)
            tooltip:SetCellColor(row, Columns["cd"..cd].col, profiledb.freecd.r, profiledb.freecd.g, profiledb.freecd.b, profiledb.freecd.a)
          end
          cd = cd+1
        end
      end 
    end
  end
end

function PT:UpdateLDB()
  local LDBcd, LDBitem
  if self.db.profile.showLDBCooldownsAvailable then 
    local availableCooldowns = 0
    for char, chardata in pairs(self.db.global.cooldowns) do
      for prof, profdata in pairs(chardata) do
        if not profdata.reset or ( profdata.reset and profdata.reset < time() ) then
          availableCooldowns = availableCooldowns + 1
        end
      end
    end
    LDBcd = strformat(self.db.profile.showLDBCooldownsText, availableCooldowns)
  end
  if self.db.profile.showLDBItemsAvailable then 
    local tmptable = {}
    local itemTexture
    for itemid, count in pairs( self.db.global.items[currentRealmChar] ) do
      if count > 0 then
        itemTexture = "|T"..select(10, GetItemInfo( itemid ) )..":16|t"
        tinsert(tmptable, count..itemTexture)
      end
    end
    if #tmptable > 0 then
      LDBitem = table.concat(tmptable, ",")
    end
  end
  PTLDB.label = ""..(LDBcd and LDBcd or "")..( ( LDBcd and LDBitem ) and "," or "")..(LDBitem and LDBitem or "")
end

-- 

function PT:OnEnable()
  self:RegisterEvent("BAG_UPDATE",               "UpdateItemCounts")
  self:RegisterEvent("PLAYER_LEVEL_UP",          "UpdatePlayer")
  self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED", "CheckForProfessionSpell")

  self:SetupClasscoloredFonts()
  self:UpdatePlayer()
  self:UpdateItemCounts()
  self:UpdateProfessionCooldowns()
end

function PT:OnDisable()
  self:UnregisterEvent("BAG_UPDATE")
  self:UnregisterEvent("PLAYER_LEVEL_UP")
  self:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
end


---- Handling Tracking

function PT:UpdatePlayer()
  local level = UnitLevel("player")
  local class = select(2,UnitClass("player"))
  self.db.global.roster[currentRealmChar] = { class = class, level = level }
end

function PT:CheckForProfessionSpell( ... )
  local spellID = select(6, ...)
  if professionCooldownSpellIDs[spellID] then
    C_Timer.After( 0.5, function() self:UpdateProfessionCooldowns() end )
  end
end


function PT:UpdateProfessionCooldowns()
  if self:GetNextDailyReset() then
    wipe( PT.db.global.cooldowns[currentRealmChar] )
    local p1, p2, skillLine
    p1, p2 = GetProfessions()
    if p1 then
      skillLine = select(7, GetProfessionInfo( p1 ))
      if professionCooldowns[skillLine] then
        self:UpdateProfessionCooldown( skillLine )
      end
    end
    if p2 then
      skillLine = select(7, GetProfessionInfo( p2 ))
      if professionCooldowns[skillLine] then
        self:UpdateProfessionCooldown( skillLine )
      end
    end
    self:UpdateLDB()
  else
    C_Timer.After( 0.5, function() self:UpdateProfessionCooldowns() end )    
  end
end

function PT:UpdateProfessionCooldown( skillLine )
  if not PT.db.global.cooldowns[currentRealmChar][skillLine] then
    PT.db.global.cooldowns[currentRealmChar][skillLine] = {}
  end

  local nextReset = self:GetNextDailyReset()
  for _, checkspell in pairs( professionCooldowns[skillLine].spells ) do
    if GetSpellCooldown(checkspell) > 0 then
      PT.db.global.cooldowns[currentRealmChar][skillLine].reset = nextReset
    end
  end
  if GetSpellCooldown( professionCooldowns[skillLine].book ) > 0 then
    PT.db.global.cooldowns[currentRealmChar][skillLine].bookReset = nextReset
  end
end

function PT:UpdateItemCounts()
  local globaldb, key, id  = self.db.global
  if not globaldb.items[currentRealmChar] then
    globaldb.items[currentRealmChar] = {}
  end
  for id, _ in pairs(globaldb.trackedItems) do
    globaldb.items[currentRealmChar][id] = GetItemCount( id, true )
    --print(currentRealmChar, id, GetItemCount( id, true ) )
  end
  self:UpdateLDB()
end

---- Database Mainteananc

function PT:GetNextDailyReset()
  -- when first logging in, we may get a bogus result from GetQuestResetTime
  -- if thats the case, then just return nil which triggers the update 
  -- functions to requeue 0.5s later
  local qrtime = GetQuestResetTime()
  if qrtime <= 0 then
    return nil
  end
  return time() + qrtime
end

